#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "ht3035.h"

/* The name this program was run with.  */
static char *program_name;

#define DEVICE "/dev/mem"
#define A460_PC104_IO_8_BASE (0xb2000000)
#define A460_PC104_IO_8_SIZE (0x10000)

static struct timespec start_time;
static volatile int exiting;
static void *ht3035_base;
#define  ht3035_readb(addr) ({*(volatile uint8_t *)((uintptr_t)ht3035_base + addr);})
#define  ht3035_writeb(addr, data) ({ \
			*(volatile uint8_t *)((uintptr_t)ht3035_base + addr) = data;})
#define  ht3035_readw(addr) ({*(volatile uint16_t *)((uintptr_t)ht3035_base + addr);})
#define  ht3035_writew(addr, data) ({ \
			*(volatile uint16_t *)((uintptr_t)ht3035_base + addr) = data;})

#define MAX(a,b) ((a) > (b) ? (a) : (b))

static struct option const long_options[] =
{
	{"velocity", required_argument, NULL, 'v'},
	{"initial", required_argument, NULL, 'i'},
	{"acceleration", required_argument, NULL, 'a'},
	{"s-curve", no_argument, NULL, 's'},
	{"pulse", required_argument, NULL, 'p'},
	{"debug", no_argument, NULL, 'd'},
	{"help", no_argument, NULL, 'h'},
	{NULL, 0, NULL, 0}
};

static void usage(int status)
{
	FILE *s;

	if (status == EXIT_SUCCESS)
		s = stdout;
	else
		s = stderr;

	fprintf(s, "Usage: %s AXES_OPTIONS CURVE_OPTIONS [OPTIONS]\n", program_name);
	fputs ("\
AXES_OPTIONS\n", s);
	fputs ("\
  -x                         control the X axis\n\
  -y                         control the Y axis\n\
  -z                         control the Z axis\n\
  -u                         control the U axis\n", s);
	fputs ("\
CURVE_OPTIONS\n", s);
	fprintf(s, "\
  -v, --velocity=NUM         drive velocity\n\
                             range: %d to %d(PPS)\n", VELOCITY_MIN, VELOCITY_MAX);
	fprintf(s, "\
  -i, --initial-velocity=NUM initial velocity\n\
                             range: %d to %d(PPS)\n",
		INIT_VELOCITY_MIN,INIT_VELOCITY_MAX);
	fprintf(s, "\
  -a, --acceleration=NUM     acceleration/deceleration\n\
                             range: %d to %d(PPS/SEC)\n",
		ACCELERATION_MIN, ACCELERATION_MAX);
	fputs("\
  -s, --s-curve              do perfect s-curve driving\n", s);
	fputs ("\
OPTIONS\n", s);
	fprintf(s, "\
  -p, --pulse=NUM            number of output pulse\n\
                             range: %d to %d\n", NR_PULSE_MIN, NR_PULSE_MAX);
	fputs ("\
  -d, --debug                display debug message\n\
  -h, --help                 display this help and exit\n", s);
	fputs ("\
EXAMPLES\n", s);
	fprintf(s, "\
  %s -xyzu -v 980 -i 980\n\
      Constant speed driving\n", program_name);
	fprintf(s, "\
  %s -xyzu -v 15000 -i 500 -a 48333\n\
      Linear acceleration driving\n", program_name);
	fprintf(s, "\
  %s -xyzu -v 40000 -i 10 -a 200000 -s\n\
      Perfect S-curve acceleration driving\n", program_name);

	exit(status);
}

static inline void _do_command(uint8_t axes, uint8_t command)
{
	/* select axes */
	ht3035_writeb(HT3035_WR0H, axes);

	/* do command */
	ht3035_writeb(HT3035_WR0L, command);

	/* must wait 250ns */
	usleep(1);
}

static void do_write_command(uint8_t axes, uint8_t command,
			     uint32_t data)
{
	ht3035_writew(HT3035_WR7, (data >> 16) & 0xffff);
	ht3035_writew(HT3035_WR6, data & 0xffff);

	_do_command(axes, command);
}

static int32_t do_read_command(uint8_t axis, uint8_t command)
{
	int32_t upper, lower;

	_do_command(axis, command);

	lower = ht3035_readw(HT3035_RR6);
	upper = ht3035_readw(HT3035_RR7);

	return ((upper << 16) | lower);
}

static void do_drive_command(uint8_t axes, uint8_t command)
{
	_do_command(axes, command);
}

static uint32_t calc_multiple(uint32_t velocity,
			      uint32_t initial_velocity,
			      uint32_t acceleration)
{
	uint32_t multiple, tmp;

	multiple = DIV_ROUND_UP(velocity, V_MAX);

	tmp = DIV_ROUND_UP(initial_velocity, SV_MAX);
	multiple = MAX(multiple, tmp);

	tmp = DIV_ROUND_UP((acceleration / 125), A_MAX);
	multiple = MAX(multiple, tmp);

	return multiple;
}

static int check_drive_info(struct drive_info *info)
{
	struct drive_param arg = info->arg;

	if (!info->axes) {
		fprintf(stderr, "at least one AXES_OPTIONS must be specified\n");
		return -1;
	}

	switch (info->curve) {
	case CURVE_CONSTANT:
		if (arg.initial_velocity < arg.velocity) {
			fprintf(stderr,
				"initial-velocity has to be equal to or higher than velocity\n");
			return -1;
		}
		break;
	case CURVE_LINEAR:
	case CURVE_PERFECT_S:
		if (arg.velocity <= arg.initial_velocity) {
			fprintf(stderr,
				"velocity has to be bigger than initial-velocity\n");
			return -1;
		}
		break;
	default:
		break;
	}

	return 0;
}

static void _set_drive_mode(uint8_t axis, int sacc)
{
	uint16_t wr3;

	do_write_command(axis, HT3035_WR0L_SET_NOP, 0);

	wr3 = ht3035_readw(HT3035_WR3);
	if (sacc)
		wr3 |= HT3035_WR3_SACC;
	else
		wr3 &= ~HT3035_WR3_SACC;

	ht3035_writew(HT3035_WR3, wr3);
}

static void set_drive_mode(struct drive_info *info)
{
	int sacc = 0;

	if (info->curve == CURVE_PERFECT_S)
		sacc = 1;

	if (info->axes & HT3035_WR0H_X_AXIS)
		_set_drive_mode(HT3035_WR0H_X_AXIS, sacc);
	if (info->axes & HT3035_WR0H_Y_AXIS)
		_set_drive_mode(HT3035_WR0H_Y_AXIS, sacc);
	if (info->axes & HT3035_WR0H_Z_AXIS)
		_set_drive_mode(HT3035_WR0H_Z_AXIS, sacc);
	if (info->axes & HT3035_WR0H_U_AXIS)
		_set_drive_mode(HT3035_WR0H_U_AXIS, sacc);
}

static uint32_t calc_k(uint32_t v, uint32_t i, uint32_t a)
{
	uint32_t ret;

	if (v <= i)
		return K_MIN;

	/*
	 * Multiple        : r
	 * Velocity        : v(PPS)
	 * Initial Velocity: i(PPS)
	 * Acceleration    : a(PPS/SEC)
	 * Jerk            : k(PPS/SEC^2)
	 * Time            : t(SEC)
	 *
	 * V, SV, A, K is register value
	 *
	 * (v - i) = 1/2 * a * t
	 * t = 2 * (v - i) / a
	 * k = 1/2 * a / t
	 *   = a^2 / (4 * (v - i))
	 *   = (A * 125 * r)^2 / (4 * (V - SV) * r)
	 * k = 62.5*10^6 / K * r
	 * K = 62.5*10^6 / k * r
	 *   = 62.5*10^6 / ((A * 125 * r)^2 / (4 * (V - SV) * r)) * r
	 *   = 4000 * (V - SV) / a^2
	 */
	ret = (4000 * (v - i)) / (a *  a);

	if (ret < K_MIN)
		return K_MIN;

	if (K_MAX < ret)
		return K_MAX;

	return ret;
}

static void convert_drive_param(struct drive_info *info)
{
	struct drive_param *arg = &info->arg;
	struct drive_param *reg = &info->reg;
	uint32_t multiple;

	multiple = calc_multiple(arg->velocity, arg->initial_velocity,
				 arg->acceleration);

	reg->multiple = MULTIPLE_TO_R(multiple);
	reg->velocity = VELOCITY_TO_V(arg->velocity, multiple);
	reg->initial_velocity = INIT_VELOCITY_TO_SV(arg->initial_velocity, multiple);
	if ((info->curve == CURVE_LINEAR) || (info->curve == CURVE_PERFECT_S))
		reg->acceleration = ACCELERATION_TO_A(arg->acceleration, multiple);
	if (info->curve == CURVE_PERFECT_S)
		reg->jerk = calc_k(reg->velocity, reg->initial_velocity,
				   reg->acceleration);
}

static void set_drive_param(struct drive_info *info)
{
	struct drive_param *reg = &info->reg;

	convert_drive_param(info);

	do_write_command(info->axes, HT3035_WR0L_SET_R, reg->multiple);
	do_write_command(info->axes, HT3035_WR0L_SET_V, reg->velocity);
	do_write_command(info->axes, HT3035_WR0L_SET_SV, reg->initial_velocity);
	if ((info->curve == CURVE_LINEAR) || (info->curve == CURVE_PERFECT_S))
		do_write_command(info->axes, HT3035_WR0L_SET_A, reg->acceleration);
	if (info->curve == CURVE_PERFECT_S)
		do_write_command(info->axes, HT3035_WR0L_SET_K, reg->jerk);
}

static inline void start_drive(struct drive_info *info)
{
	do_write_command(info->axes, HT3035_WR0L_SET_P,
			 NR_PULSE_TO_P(info->nr_pulse));

	if (info->nr_pulse)
		do_drive_command(info->axes, HT3035_WR0L_DRV_PF);
	else
		do_drive_command(info->axes, HT3035_WR0L_DRV_PC);
}

static inline void stop_drive(struct drive_info *info)
{
	do_drive_command(info->axes, HT3035_WR0L_DRV_SS);
}

static int is_driving(struct drive_info *info)
{
	return !!(ht3035_readw(HT3035_RR0) & info->axes);
}

static void init_ht3035(struct drive_info *info)
{
	/* reset MCX304 */
	ht3035_writeb(HT3035_WR0H, HT3035_WR0H_RESET);
	ht3035_writeb(HT3035_WR0L, HT3035_WR0L_SET_NOP);
	/* must wait 875ns */
	usleep(1);

	/* clear logical position counter */
	do_write_command(info->axes, HT3035_WR0L_SET_LP, 0);
}

static inline void read_counter(int32_t *x, int32_t *y, int32_t *z, int32_t *u)
{
	*x = do_read_command(HT3035_WR0H_X_AXIS, HT3035_WR0L_GET_LP);
	*y = do_read_command(HT3035_WR0H_Y_AXIS, HT3035_WR0L_GET_LP);
	*z = do_read_command(HT3035_WR0H_Z_AXIS, HT3035_WR0L_GET_LP);
	*u = do_read_command(HT3035_WR0H_U_AXIS, HT3035_WR0L_GET_LP);
}

static inline void print_header(void)
{
	printf("--- time --- ---- X ---- ---- Y ---- ---- Z ---- ---- U ----\n");
}

static inline void tssub(struct timespec *out, struct timespec *in)
{
	if ((out->tv_nsec -= in->tv_nsec) < 0) {
		--out->tv_sec;
		out->tv_nsec += 1000000000;
	}
	out->tv_sec -= in->tv_sec;
}

static void print_counter(void)
{
	static int32_t xpos, ypos, zpos, upos = -1;
	int32_t nxpos, nypos, nzpos, nupos;
	struct timespec ts;

	clock_gettime(CLOCK_MONOTONIC, &ts);
	tssub(&ts, &start_time);

	read_counter(&nxpos, &nypos, &nzpos, &nupos);

	if ((xpos != nxpos) || (ypos != nypos) || (zpos != nzpos) || (upos != nupos))
		printf("%5ld.%06ld %11d %11d %11d %11d\n", ts.tv_sec, ts.tv_nsec / 1000,
		       xpos = nxpos, ypos = nypos, zpos = nzpos, upos = nupos);
}

static int xmmap(void)
{
	int fd;

	fd = open(DEVICE, O_RDWR | O_SYNC);
	if (fd < 0) {
		perror("open");
		return -1;
	}

	ht3035_base = mmap(0x0, A460_PC104_IO_8_SIZE,
			   PROT_READ | PROT_WRITE, MAP_SHARED,
			   fd, A460_PC104_IO_8_BASE);
	if (ht3035_base == MAP_FAILED) {
		perror("mmap");
		close(fd);
		return -1;
	}

	close(fd);

	return 0;
}

static void xmunmap(void)
{
	munmap(ht3035_base, A460_PC104_IO_8_SIZE);
}

static int xstrtoul(const char *str, unsigned long *val)
{
	char *endptr;

	errno = 0;
	*val = strtol(str, &endptr, 0);

	if (errno || *endptr)
		return -1;

	return 0;
}

static void xsigaction(int signo, void (*handler)(int))
{
	struct sigaction sa;

	memset(&sa, 0, sizeof(sa));

	sa.sa_handler = (void (*)(int))handler;
	sigaction(signo, &sa, NULL);
}

static void sigexit(__attribute__((unused)) int signo)
{
	exiting = 1;
}

int main(int argc, char *argv[])
{
	struct drive_info info;
	struct drive_param *arg = &info.arg;
	unsigned long tmp;
	int debug = 0;
	int ret;

	program_name = argv[0];
	memset(&info, 0, sizeof(info));

	while (1) {
		int c = getopt_long(argc, argv, "xyzuv:i:a:sp:dh",
				    long_options, NULL);
		if (c == -1)
			break;

		switch (c) {
		case 'x':
			info.axes |= HT3035_WR0H_X_AXIS;
			break;
		case 'y':
			info.axes |= HT3035_WR0H_Y_AXIS;
			break;
		case 'z':
			info.axes |= HT3035_WR0H_Z_AXIS;
			break;
		case 'u':
			info.axes |= HT3035_WR0H_U_AXIS;
			break;
		case 'v':
			ret = xstrtoul(optarg, &tmp);
			if (ret < 0) {
				fprintf(stderr, "invalid drive velocity: %s\n",
					optarg);
				return EXIT_FAILURE;
			}
			if ((tmp < VELOCITY_MIN) || (VELOCITY_MAX < tmp)) {
				fprintf(stderr, "out of range(velocity=%ld)\n", tmp);
				return -1;
			}
			arg->velocity = tmp;
			info.curve |= CURVE_V;
			break;
		case 'i':
			ret = xstrtoul(optarg, &tmp);
			if (ret < 0) {
				fprintf(stderr, "invalid initial-velocity: %s\n",
					optarg);
				return EXIT_FAILURE;
			}
			if ((tmp < INIT_VELOCITY_MIN) || (INIT_VELOCITY_MAX < tmp)) {
				fprintf(stderr,
					"out of range(initial-velocity=%ld)\n", tmp);
				return -1;
			}
			arg->initial_velocity = tmp;
			info.curve |= CURVE_I;
			break;
		case 'a':
			ret = xstrtoul(optarg, &tmp);
			if (ret < 0) {
				fprintf(stderr, "invalid acceleration: %s\n",
					optarg);
				return EXIT_FAILURE;
			}
			if ((tmp < ACCELERATION_MIN) || (ACCELERATION_MAX < tmp)) {
				fprintf(stderr,
					"out of range(acceleration=%ld)\n", tmp);
				return -1;
			}
			arg->acceleration = tmp;
			info.curve |= CURVE_A;
			break;
		case 's':
			info.curve |= CURVE_S;
			break;
		case 'p':
			ret = xstrtoul(optarg, &tmp);
			if (ret < 0) {
				fprintf(stderr, "invalid nr_pulse: %s\n", optarg);
				return EXIT_FAILURE;
			}
			if (NR_PULSE_MAX < tmp) {
				fprintf(stderr, "out of range(nr_pulse=%ld)\n", tmp);
				return -1;
			}
			info.nr_pulse = tmp;
			break;
		case 'd':
			debug = 1;
			break;
		case 'h':
			usage(EXIT_SUCCESS);
			break;
		case '?':
		default:
			usage(EXIT_FAILURE);
		}
	}

	ret = check_drive_info(&info);
	if (ret)
		usage(EXIT_FAILURE);

	ret = xmmap();
	if (ret)
		return EXIT_FAILURE;
	ret = atexit(xmunmap);
	if (ret)
		return EXIT_FAILURE;

	init_ht3035(&info);

	set_drive_mode(&info);
	set_drive_param(&info);

	exiting = 0;
	xsigaction(SIGINT, sigexit);
	start_drive(&info);

	printf("Press Ctrl-c to stop.\n");

	if (debug) {
		print_header();
		clock_gettime(CLOCK_MONOTONIC, &start_time);
	}

	while (!exiting) {
		int drive;

		usleep(10000);
		if (debug)
			print_counter();
		if (info.nr_pulse) {
			drive = is_driving(&info);
			if (!drive)
				break;
		}
	}

	if (exiting)
		stop_drive(&info);

	return EXIT_SUCCESS;
}

